@@ -8,13 +8,20 @@ from sklearn.metrics import pairwise_distances_argmin_min
 from sklearn.preprocessing import minmax_scale
 from keras_preprocessing.image import load_img
 from collections import defaultdict, Counter
+from pointgrid import align_points_to_grid
 from distutils.dir_util import copy_tree
 from iiif_downloader import Manifest
 from sklearn.cluster import KMeans
+from rasterfairy import coonswarp
+import matplotlib.pyplot as plt
 from keras.models import Model
+from scipy.stats import kde
+from hdbscan import HDBSCAN
 from hashlib import sha224
 import keras.backend as K
+import tensorflow as tf
 from umap import UMAP
+import multiprocessing
 import pkg_resources
 import rasterfairy
 import numpy as np
@@ -23,8 +30,10 @@ import functools
 import itertools
 import datetime
 import argparse
+import random
 import shutil
 import glob2
+import copy
 import uuid
 import math
 import gzip
@@ -33,24 +42,38 @@ import sys
 import csv
 import os
 
+try:
+  from MulticoreTSNE import MulticoreTSNE as TSNE
+except:
+  from sklearn.manifold import TSNE
+
 try:
   from urllib.parse import unquote # python 3
 except:
   from urllib import unquote # python 2
 
+'''
+NB: Keras Image class objects return image.size as w,h
+    Numpy array representations of images return image.shape as h,w,c
+'''
 
 config = {
   'images': None,
   'metadata': None,
   'out_dir': 'output',
-  'use_cache': False,
-  'use_gzip': False,
+  'use_cache': True,
   'encoding': 'utf8',
-  'n_clusters': 20,
+  'min_cluster_size': 20,
   'atlas_size': 2048,
+  'cell_size': 32,
   'lod_cell_height': 128,
-  'atlas_cell_height': 32,
+  'n_neighbors': 6,
+  'min_dist': 0.001,
+  'metric': 'correlation',
+  'pointgrid_fill': 0.05,
   'square_cells': False,
+  'gzip': False,
+  'plot_id': str(uuid.uuid1()),
 }
 
 ##
@@ -62,7 +85,7 @@ def process_images(**kwargs):
   copy_web_assets(**kwargs)
   kwargs['out_dir'] = join(kwargs['out_dir'], 'data')
   kwargs['image_paths'] = filter_images(**kwargs)
-  kwargs['atlas_positions'] = get_atlas_positions(**kwargs)
+  kwargs['atlas_dir'] = get_atlas_data(**kwargs)
   get_manifest(**kwargs)
   write_images(**kwargs)
   print(' * done!')
@@ -97,64 +120,87 @@ def write_images(**kwargs):
 
 def get_manifest(**kwargs):
   '''Create and return the base object for the manifest output file'''
-  atlas_ids = set([i['idx'] for i in kwargs['atlas_positions']])
+  # load the atlas data
+  atlas_data = json.load(open(join(kwargs['atlas_dir'], 'atlas_positions.json')))
+  # store each cell's size and atlas position
+  atlas_ids = set([i['idx'] for i in atlas_data])
   sizes = [[] for _ in atlas_ids]
   pos = [[] for _ in atlas_ids]
-  for idx, i in enumerate(kwargs['atlas_positions']):
+  for idx, i in enumerate(atlas_data):
     sizes[ i['idx'] ].append([ i['w'], i['h'] ])
     pos[ i['idx'] ].append([ i['x'], i['y'] ])
-  # create base metadata for manifest
+  # obtain the paths to each layout's JSON positions
+  layouts = get_layouts(**kwargs)
+  # create a heightmap for each layout
+  for i in layouts:
+    for j in layouts[i]:
+      get_heightmap(layouts[i][j], i + '-' + j, **kwargs)
+  # create manifest json
   manifest = {
+    'layouts': layouts,
+    'initial_layout': 'grid',
+    'point_size': 1 / math.ceil( len(kwargs['image_paths'])**(1/2) ),
+    'imagelist': get_path('imagelists', 'imagelist', **kwargs),
+    'atlas_dir': kwargs['atlas_dir'],
+    'config': {
+      'sizes': {
+        'atlas': kwargs['atlas_size'],
+        'cell': kwargs['cell_size'],
+        'lod': kwargs['lod_cell_height'],
+      },
+    },
+    'metadata': True if kwargs['metadata'] else False,
+    'centroids': get_centroids(vecs=read_json(layouts['umap']['layout'], **kwargs), **kwargs),
+    'creation_date': datetime.datetime.today().strftime('%d-%B-%Y-%H:%M:%S'),
+    'version': get_version(),
+  }
+  path = get_path('manifests', 'manifest', **kwargs)
+  write_json(path, manifest, **kwargs)
+  path = get_path(None, 'manifest', add_hash=False, **kwargs)
+  write_json(path, manifest, **kwargs)
+  # create images json
+  imagelist = {
     'cell_sizes': sizes,
+    'images': [clean_filename(i) for i in kwargs['image_paths']],
     'atlas': {
       'count': len(atlas_ids),
       'positions': pos,
     },
-    'layouts': {},
-    'config': {
-      'sizes': {
-        'atlas': kwargs['atlas_size'],
-        'cell': kwargs['atlas_cell_height'],
-        'lod': kwargs['lod_cell_height'],
-      }
-    },
-    'images': [clean_filename(i) for i in kwargs['image_paths']],
-    'metadata': True if kwargs['metadata'] else False,
-    'creation_date': datetime.datetime.today().strftime('%d-%B-%Y-%H:%M:%S'),
-    'version': get_version(),
   }
-  # compute centroids for each layout and add to manifest
-  for label, position_path in get_positions(**kwargs).items():
-    out_dir = join(kwargs['out_dir'], 'centroids')
-    if not exists(out_dir): os.makedirs(out_dir)
-    centroid_path = join(out_dir, hash(**kwargs) + '.json')
-    if not os.path.exists(centroid_path):
-      vecs = json.load(open(position_path))
-      centroids = get_centroids(vecs=vecs, **kwargs)
-      with open(centroid_path, 'w') as out: json.dump(centroids, out)
-    manifest['layouts'][label] = {
-      'positions': '/' + position_path,
-      'centroids': '/' + centroid_path,
-    }
-  out_dir = join(kwargs['out_dir'], 'manifests')
-  if not exists(out_dir): os.makedirs(out_dir)
-  path = join(out_dir, hash(**kwargs) + '.json')
-  with open(path, 'w') as out: json.dump(manifest, out)
-  path = join(kwargs['out_dir'], 'manifest.json')
-  with open(path, 'w') as out: json.dump(manifest, out)
-  return manifest
+  write_json(manifest['imagelist'], imagelist)
 
 
 def filter_images(**kwargs):
   '''Main method for filtering images given user metadata (if provided)'''
-  image_paths = [i.path for i in stream_images(image_paths=get_image_paths(**kwargs))]
+  image_paths = []
+  for i in stream_images(image_paths=get_image_paths(**kwargs)):
+    # get image height and width
+    w,h = i.original.size
+    # remove images with 0 height or width when resized to lod height
+    if (h == 0) or (w == 0):
+      print(' * skipping {} because it contains 0 height or width'.format(i.path))
+      continue
+    # remove images that have 0 height or width when resized
+    try:
+      resized = i.resize_to_max(kwargs['lod_cell_height'])
+    except ValueError:
+      print(' * skipping {} because it contains 0 height or width when resized'.format(i.path))
+      continue
+    # remove images that are too wide for the atlas
+    if (w/h) > (kwargs['atlas_size']/kwargs['cell_size']):
+      print(' * skipping {} because its dimensions are oblong'.format(i.path))
+      continue
+    image_paths.append(i.path)
+  # handle the case user provided no metadata
   if not kwargs.get('metadata', False): return image_paths
+  # handle csv metadata
   l = []
   if kwargs['metadata'].endswith('.csv'):
     headers = ['filename', 'tags', 'description', 'permalink']
     with open(kwargs['metadata']) as f:
       for i in list(csv.reader(f)):
         l.append({headers[j]: i[j] for j,_ in enumerate(headers)})
+  # handle json metadata
   else:
     for i in glob2.glob(kwargs['metadata']):
       with open(i) as f:
@@ -174,8 +220,13 @@ def filter_images(**kwargs):
 
 
 def get_image_paths(**kwargs):
+  '''Called once to provide a list of image paths--handles IIIF manifest input'''
   # handle case where --images points to iiif manifest
   image_paths = None
+  if not kwargs['images']:
+    print('\nError: please provide an images argument, e.g.:')
+    print('pixplot --images "cat_pictures/*.jpg"\n')
+    sys.exit()
   if os.path.exists(kwargs['images']):
     with open(kwargs['images']) as f:
       f = [i.strip() for i in f.read().split('\n') if i.strip()]
@@ -186,15 +237,12 @@ def get_image_paths(**kwargs):
   if not image_paths:
     image_paths = sorted(glob2.glob(kwargs['images']))
   if not image_paths:
-    print('\nError: No input images were found. Please check your --images glob')
-    sys.exit()
-  # validate the provided images are > n clusters requested
-  n_images = len(image_paths)
-  n_clusters = kwargs['n_clusters']
-  if n_images <= n_clusters:
-    print('\nError: n_clusters must be < input image count ')
-    print('Found {} images and {} clusters were requested'.format(n_images, n_clusters))
+    print('\nError: No input images were found. Please check your --images glob\n')
     sys.exit()
+  # optional shuffle that mutates image_paths
+  if kwargs['shuffle']:
+    print(' * shuffling input images')
+    random.shuffle(image_paths)
   return image_paths
 
 
@@ -216,51 +264,57 @@ def write_metadata(**kwargs):
     tags = [j.strip() for j in i['tags'].split('|')]
     i['tags'] = tags
     for j in tags: d[ '__'.join(j.split()) ].append(filename)
-    out_path = os.path.join(out_dir, 'file', filename + '.json')
-    with open(out_path, 'w') as out: json.dump(i, out)
+    write_json(os.path.join(out_dir, 'file', filename + '.json'), i, **kwargs)
   filters = [{'filter_name': 'select', 'filter_values': list(d.keys())}]
-  with open(os.path.join(out_dir, 'filters', 'filters.json'), 'w') as out:
-    json.dump(filters, out)
+  write_json(os.path.join(out_dir, 'filters', 'filters.json'), filters, **kwargs)
   # create the options
   for i in d:
-    with open(os.path.join(out_dir, 'options', i + '.json'), 'w') as out:
-      json.dump(d[i], out)
+    write_json(os.path.join(out_dir, 'options', i + '.json'), d[i], **kwargs)
 
 
-def get_atlas_positions(**kwargs):
+def get_atlas_data(**kwargs):
   '''
   Generate and save to disk all atlases to be used for this visualization
   If square, center each cell in an nxn square, else use uniform height
   '''
+  # if the atlas files already exist, load from cache
+  out_dir = os.path.join(kwargs['out_dir'], 'atlases', kwargs['plot_id'])
+  if os.path.exists(out_dir) and kwargs['use_cache']:
+    print(' * loading saved atlas data')
+    return out_dir
+  if not os.path.exists(out_dir):
+    os.makedirs(out_dir)
+  # else create the atlas images and store the positions of cells in atlases
   print(' * creating atlas files')
   n = 0 # number of atlases
   x = 0 # x pos in atlas
   y = 0 # y pos in atlas
-  atlas_positions = [] # l[cell_idx] = atlas data
+  positions = [] # l[cell_idx] = atlas data
   atlas = np.zeros((kwargs['atlas_size'], kwargs['atlas_size'], 3))
   for idx, i in enumerate(stream_images(**kwargs)):
-    lod_data = i.resize_to_max(kwargs['lod_cell_height'])
-    h,w,_ = lod_data.shape # h,w,colors in lod-cell sized image `i`
     if kwargs['square_cells']:
-      atl_data = i.resize_to_square(kwargs['atlas_cell_height'])
+      cell_data = i.resize_to_square(kwargs['cell_size'])
     else:
-      atl_data = i.resize_to_height(kwargs['atlas_cell_height'])
-    _,v,_ = atl_data.shape
+      cell_data = i.resize_to_height(kwargs['cell_size'])
+    _, v, _ = cell_data.shape
     appendable = False
-    if x + v < kwargs['atlas_size']:
+    if (x + v) <= kwargs['atlas_size']:
       appendable = True
-    elif y + kwargs['atlas_cell_height'] < kwargs['atlas_size']:
-      y += kwargs['atlas_cell_height']
+    elif (y + (2*kwargs['cell_size'])) <= kwargs['atlas_size']:
+      y += kwargs['cell_size']
       x = 0
       appendable = True
     if not appendable:
-      save_atlas(atlas, n, **kwargs)
+      save_atlas(atlas, out_dir, n)
       n += 1
       atlas = np.zeros((kwargs['atlas_size'], kwargs['atlas_size'], 3))
       x = 0
       y = 0
-    atlas[y:y+kwargs['atlas_cell_height'], x:x+v] = atl_data
-    atlas_positions.append({
+    atlas[y:y+kwargs['cell_size'], x:x+v] = cell_data
+    # find the size of the cell in the lod canvas
+    lod_data = i.resize_to_max(kwargs['lod_cell_height'])
+    h,w,_ = lod_data.shape # h,w,colors in lod-cell sized image `i`
+    positions.append({
       'idx': n, # atlas idx
       'x': x, # x offset of cell in atlas
       'y': y, # y offset of cell in atlas
@@ -268,28 +322,37 @@ def get_atlas_positions(**kwargs):
       'h': h, # h of cell at lod size
     })
     x += v
-  save_atlas(atlas, n, **kwargs)
-  return atlas_positions
+  save_atlas(atlas, out_dir, n)
+  out_path = os.path.join(out_dir, 'atlas_positions.json')
+  with open(out_path, 'w') as out:
+    json.dump(positions, out)
+  return out_dir
 
 
-def save_atlas(*args, **kwargs):
+def save_atlas(atlas, out_dir, n):
   '''Save an atlas to disk'''
-  data, n = args
-  out_dir = join(kwargs['out_dir'], 'atlases')
-  if not exists(out_dir): os.makedirs(out_dir)
-  save_img(join(out_dir, 'atlas-{}.jpg'.format(n)), data)
+  out_path = join(out_dir, 'atlas-{}.jpg'.format(n))
+  save_img(out_path, atlas)
 
 
-def get_positions(*args, **kwargs):
+def get_layouts(*args, **kwargs):
   '''Get the image positions in each projection'''
   vecs = vectorize_images(**kwargs)
   umap = get_umap_projection(vecs=vecs, **kwargs)
-  rasterfairy = get_rasterfairy_projection(umap=umap, **kwargs)
+  raster = get_rasterfairy_projection(umap=umap, **kwargs)
   grid = get_grid_projection(**kwargs)
+  umap_grid = get_pointgrid_projection(umap, 'umap', **kwargs)
   return {
-    'umap': umap,
-    'grid': grid,
-    'rasterfairy': rasterfairy,
+    'umap': {
+      'layout': umap,
+      'jittered': umap_grid,
+    },
+    'grid': {
+      'layout': grid,
+    },
+    'rasterfairy': {
+      'layout': raster,
+    },
   }
 
 
@@ -304,7 +367,7 @@ def vectorize_images(**kwargs):
   vecs = []
   for idx, i in enumerate(stream_images(**kwargs)):
     vector_path = os.path.join(vector_dir, os.path.basename(i.path) + '.npy')
-    if os.path.exists(vector_path):
+    if os.path.exists(vector_path) and kwargs['use_cache']:
       vec = np.load(vector_path)
     else:
       im = preprocess_input( img_to_array( i.original.resize((299,299)) ) )
@@ -318,34 +381,45 @@ def vectorize_images(**kwargs):
 def get_umap_projection(**kwargs):
   '''Get the x,y positions of images passed through a umap projection'''
   print(' * creating UMAP layout')
-  out_dir = join(kwargs['out_dir'], 'layouts')
-  out_path = join(out_dir, 'umap-{}.json'.format(hash(**kwargs)))
-  if os.path.exists(out_path): return out_path
-  model = UMAP(n_neighbors=25, min_dist=0.5, metric='correlation')
+  out_path = get_path('layouts', 'umap', **kwargs)
+  if os.path.exists(out_path) and kwargs['use_cache']: return out_path
+  model = UMAP(n_neighbors=kwargs['n_neighbors'],
+    min_dist=kwargs['min_dist'],
+    metric=kwargs['metric'])
   z = model.fit_transform(kwargs['vecs'])
-  path = write_json(out_path, z, **kwargs)
-  return path
+  return write_layout(out_path, z, **kwargs)
+
+
+def get_tsne_projection(**kwargs):
+  '''Get the x,y positions of images passed through a TSNE projection'''
+  print(' * creating TSNE layout with ' + str(multiprocessing.cpu_count()) + ' cores...')
+  out_path = get_path('layouts', 'tsne', **kwargs)
+  if os.path.exists(out_path) and kwargs['use_cache']: return out_path
+  model = TSNE(perplexity=kwargs.get('perplexity', 2),n_jobs=multiprocessing.cpu_count())
+  z = model.fit_transform(kwargs['vecs'])
+  return write_layout(out_path, z, **kwargs)
 
 
 def get_rasterfairy_projection(**kwargs):
   '''Get the x, y position of images passed through a rasterfairy projection'''
   print(' * creating rasterfairy layout')
-  out_dir = join(kwargs['out_dir'], 'layouts')
-  out_path = join(out_dir, 'rasterfairy-{}.json'.format(hash(**kwargs)))
-  if os.path.exists(out_path): return out_path
-  umap = np.array(json.load(open(kwargs['umap'])))
-  umap = (umap + 1) * 100 # make positive and upscale
+  out_path = get_path('layouts', 'rasterfairy', **kwargs)
+  if os.path.exists(out_path) and kwargs['use_cache']: return out_path
+  umap = np.array(read_json(kwargs['umap'], **kwargs))
+  umap = (umap + 1)/2 # scale 0:1
+  umap = coonswarp.rectifyCloud(umap, # stretch the distribution
+    perimeterSubdivisionSteps=4,
+    autoPerimeterOffset=False,
+    paddingScale=1.05)
   pos = rasterfairy.transformPointCloud2D(umap)[0]
-  path = write_json(out_path, pos, **kwargs)
-  return path
+  return write_layout(out_path, pos, **kwargs)
 
 
 def get_grid_projection(**kwargs):
   '''Get the x,y positions of images in a grid projection'''
   print(' * creating grid layout')
-  out_dir = join(kwargs['out_dir'], 'layouts')
-  out_path = join(out_dir, 'grid-{}.json'.format(hash(**kwargs)))
-  if os.path.exists(out_path): return out_path
+  out_path = get_path('layouts', 'grid', **kwargs)
+  if os.path.exists(out_path) and kwargs['use_cache']: return out_path
   paths = kwargs['image_paths']
   n = math.ceil(len(paths)**(1/2))
   l = [] # positions
@@ -354,8 +428,17 @@ def get_grid_projection(**kwargs):
     y = math.floor(i/n)
     l.append([x, y])
   z = np.array(l)
-  path = write_json(out_path, z, **kwargs)
-  return path
+  return write_layout(out_path, z, **kwargs)
+
+
+def get_pointgrid_projection(path, label, **kwargs):
+  '''Gridify the positions in `path` and return the path to this new layout'''
+  print(' * creating {} pointgrid'.format(label))
+  out_path = get_path('layouts', label + '-jittered', **kwargs)
+  if os.path.exists(out_path) and kwargs['use_cache']: return out_path
+  arr = np.array(read_json(path, **kwargs))
+  z = align_points_to_grid(arr)
+  return write_layout(out_path, z, **kwargs)
 
 
 def add_z_dim(X, val=0.001):
@@ -367,34 +450,75 @@ def add_z_dim(X, val=0.001):
   return X.tolist()
 
 
-def write_json(path, obj, precision=4, sub_dir='layouts', **kwargs):
-  '''Write json object `o` to disk and return the path to that file'''
-  obj = minmax_scale(obj)
+def get_path(*args, **kwargs):
+  '''Return the path to a JSON file with conditional gz extension'''
+  sub_dir, filename = args
+  out_dir = join(kwargs['out_dir'], sub_dir) if sub_dir else kwargs['out_dir']
+  if kwargs.get('add_hash', True):
+    filename += '-' + kwargs['plot_id']
+  path = join(out_dir, filename + '.json')
+  return path + '.gz' if kwargs.get('gzip', False) else path
+
+
+def write_layout(path, obj, **kwargs):
+  '''Write layout json `obj` to disk and return the path to the saved file'''
+  obj = (minmax_scale(obj)-0.5)*2 # scale -1:1
+  obj = [[round(float(j), 5) for j in i] for i in obj]
+  return write_json(path, obj, **kwargs)
+
+
+def write_json(path, obj, **kwargs):
+  '''Write json object `obj` to disk and return the path to that file'''
   out_dir, filename = os.path.split(path)
   if not os.path.exists(out_dir): os.makedirs(out_dir)
-  if precision: obj = [[round(float(j), precision) for j in i] for i in obj]
-  with open(path, 'w') as out: json.dump(obj, out)
-  return path
+  if kwargs.get('gzip', False):
+    with gzip.GzipFile(path, 'w') as out:
+      out.write(json.dumps(obj).encode(kwargs['encoding']))
+    return path
+  else:
+    with open(path, 'w') as out:
+      json.dump(obj, out)
+    return path
+
+
+def read_json(path, **kwargs):
+  '''Read and return the json object written by the current process at `path`'''
+  if kwargs.get('gzip', False):
+    with gzip.GzipFile(path, 'r') as f:
+      return json.loads(f.read().decode(kwargs['encoding']))
+  with open(path) as f:
+    return json.load(f)
 
 
 def get_centroids(**kwargs):
-  '''Return the K nearest neighbor centroids for input vectors'''
-  z = KMeans(n_clusters=kwargs['n_clusters']).fit(kwargs['vecs'])
-  centroids = z.cluster_centers_
+  '''Return the stable clusters from the condensed tree of connected components from the density graph'''
+  print(' * HDBSCAN clustering data with ' + str(multiprocessing.cpu_count()) + ' cores...')
+  config = {
+    'min_cluster_size': kwargs['min_cluster_size'],
+    'cluster_selection_epsilon': 0.01,
+    'min_samples': 1,
+    'core_dist_n_jobs': multiprocessing.cpu_count(),
+  }
+  z = HDBSCAN(**config).fit(kwargs['vecs'])
+  # find the centroids for each cluster
+  d = defaultdict(list)
+  for idx, i in enumerate(z.labels_):
+    d[i].append(kwargs['vecs'][idx])
+  centroids = []
+  for i in d:
+    x, y = np.array(d[i]).T
+    centroids.append(np.array([np.sum(x)/len(x), np.sum(y)/len(y)]))
   closest, _ = pairwise_distances_argmin_min(centroids, kwargs['vecs'])
+  closest = set(closest)
+  print(' * found', len(closest), 'clusters')
   paths = [kwargs['image_paths'][i] for i in closest]
-  return [{
+  data = [{
     'img': clean_filename(paths[idx]),
     'label': 'Cluster {}'.format(idx+1),
     'idx': int(i),
-    'position': centroids[idx].tolist(),
   } for idx,i in enumerate(closest)]
-
-
-def hash(**kwargs):
-  '''Hash `args` into a string and return that string. Overloads hash()'''
-  return str(uuid.uuid1())
-  return sha224( u''.join([str(i) for i in kwargs]).encode('utf8')).hexdigest()
+  # save the centroids to disk and return the path to the saved json
+  return write_json(get_path('centroids', 'centroid', **kwargs), data, **kwargs)
 
 
 def copy_web_assets(**kwargs):
@@ -409,21 +533,32 @@ def copy_web_assets(**kwargs):
       f = f.read().replace('VERSION_NUMBER', get_version())
       with open(path, 'w') as out:
         out.write(f)
+  if kwargs['copy_web_only']: sys.exit()
 
 
-def parse():
-  '''Read command line args and begin data processing'''
-  description = 'Generate the data required to create a PixPlot viewer'
-  parser = argparse.ArgumentParser(description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
-  parser.add_argument('--images', type=str, default=config['images'], help='path to a glob of images to process', required=True)
-  parser.add_argument('--metadata', type=str, default=config['metadata'], help='path to a csv or glob of JSON files with image metadata (see readme for format)', required=False)
-  parser.add_argument('--use_cache', type=bool, default=config['use_cache'], help='given inputs identical to prior inputs, load outputs from cache', required=False)
-  parser.add_argument('--use_gzip', type=bool, default=config['use_gzip'], help='save outputs with gzip compression', required=False)
-  parser.add_argument('--encoding', type=str, default=config['encoding'], help='the encoding of input metadata', required=False)
-  parser.add_argument('--n_clusters', type=int, default=config['n_clusters'], help='the number of clusters to identify', required=False)
-  parser.add_argument('--out_dir', type=str, default=config['out_dir'], help='the directory to which outputs will be saved', required=False)
-  config.update(vars(parser.parse_args()))
-  process_images(**config)
+def get_version():
+  '''Return the version of pixplot installed'''
+  return pkg_resources.get_distribution('pixplot').version
+
+
+def get_heightmap(path, label, **kwargs):
+  '''Create a heightmap using the distribution of points stored at `path`'''
+  X = np.array(read_json(path, **kwargs))
+  # create kernel density estimate of distribution X
+  nbins = 200
+  x, y = X.T
+  xi, yi = np.mgrid[x.min():x.max():nbins*1j, y.min():y.max():nbins*1j]
+  zi = kde.gaussian_kde(X.T)(np.vstack([xi.flatten(), yi.flatten()]))
+  # create the plot
+  fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(5,5))
+  fig.subplots_adjust(0,0,1,1)
+  plt.pcolormesh(xi, yi, zi.reshape(xi.shape), shading='gouraud', cmap=plt.cm.gray)
+  plt.axis('off')
+  # save the plot
+  out_dir = os.path.join(kwargs['out_dir'], 'heightmaps')
+  if not os.path.exists(out_dir): os.makedirs(out_dir)
+  out_path = os.path.join(out_dir, label + '-heightmap.png')
+  plt.savefig(out_path, pad_inches=0)
 
 
 class Image:
@@ -444,7 +579,11 @@ class Image:
     Resize self.original into an image with height h and proportional width
     '''
     w,h = self.original.size
-    size = (int(w/h*height), height)
+    if (w/h*height) < 1:
+      resizedwidth = 1
+    else:
+      resizedwidth =  int(w/h*height)
+    size = (resizedwidth, height)
     return img_to_array(self.original.resize(size))
 
   def resize_to_square(self, n, center=False):
@@ -453,17 +592,41 @@ class Image:
     if center, center the colored pixels in the square, else left align
     '''
     a = self.resize_to_max(n)
-    w,h,c = a.shape
+    h,w,c = a.shape
     pad_lr = int((n-w)/2) # left right pad
     pad_tb = int((n-h)/2) # top bottom pad
     b = np.zeros((n,n,3))
-    if center: b[ pad_lr:pad_lr+w, pad_tb:pad_tb+h, : ] = a
-    else: b[:w,:h,:] = a
+    if center: b[ pad_tb:pad_tb+h, pad_lr:pad_lr+w, : ] = a
+    else: b[:h, :w, :] = a
     return b
 
-def get_version():
-  '''Return the version of pixplot installed'''
-  return pkg_resources.get_distribution('pixplot').version
+
+##
+# Entry Point
+##
+
+
+def parse():
+  '''Read command line args and begin data processing'''
+  description = 'Generate the data required to create a PixPlot viewer'
+  parser = argparse.ArgumentParser(description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
+  parser.add_argument('--images', type=str, default=config['images'], help='path to a glob of images to process', required=False)
+  parser.add_argument('--metadata', type=str, default=config['metadata'], help='path to a csv or glob of JSON files with image metadata (see readme for format)', required=False)
+  parser.add_argument('--use_cache', type=bool, default=config['use_cache'], help='given inputs identical to prior inputs, load outputs from cache', required=False)
+  parser.add_argument('--encoding', type=str, default=config['encoding'], help='the encoding of input metadata', required=False)
+  parser.add_argument('--min_cluster_size', type=int, default=config['min_cluster_size'], help='the minimum number of images in a cluster', required=False)
+  parser.add_argument('--out_dir', type=str, default=config['out_dir'], help='the directory to which outputs will be saved', required=False)
+  parser.add_argument('--cell_size', type=int, default=config['cell_size'], help='the size of atlas cells in px', required=False)
+  parser.add_argument('--n_neighbors', type=int, default=config['n_neighbors'], help='the n_neighbors argument for UMAP')
+  parser.add_argument('--min_dist', type=float, default=config['min_dist'], help='the min_dist argument for umap')
+  parser.add_argument('--metric', type=str, default=config['metric'], help='the metric argument for umap')
+  parser.add_argument('--pointgrid_fill', type=float, default=config['pointgrid_fill'], help='float 0:1 that determines sparsity of jittered distributions (lower means more sparse)')
+  parser.add_argument('--copy_web_only', action='store_true', help='update ./output/web without reprocessing data')
+  parser.add_argument('--gzip', action='store_true', help='save outputs with gzip compression')
+  parser.add_argument('--shuffle', action='store_true', help='shuffle the input images before data processing begins')
+  parser.add_argument('--plot_id', type=str, default=config['plot_id'], help='unique id for a plot; useful for resuming processing on a started plot')
+  config.update(vars(parser.parse_args()))
+  process_images(**config)
 
 if __name__ == '__main__':
-  parse()
\ No newline at end of file
+  parse()
